@@ -15,9 +15,11 @@ running multiple build commands in parallel, confusion can occur).
 import argparse
 import logging
 import sys
+import time
 
 from tools import shared
 from tools import system_libs
+from tools.settings import settings
 import emscripten
 
 
@@ -29,14 +31,14 @@ SYSTEM_TASKS = list(SYSTEM_LIBRARIES.keys())
 SYSTEM_TASKS += ['struct_info']
 
 # Minimal subset of SYSTEM_TASKS used by CI systems to build enough to useful
-# TODO Re-add 'except' versions of libc++abi, libc++, and libunwind after the
-# new EH implementation is stablized
 MINIMAL_TASKS = [
     'libcompiler_rt',
     'libc',
     'libc++abi',
+    'libc++abi-except',
     'libc++abi-noexcept',
     'libc++',
+    'libc++-except',
     'libc++-noexcept',
     'libal',
     'libdlmalloc',
@@ -46,13 +48,15 @@ MINIMAL_TASKS = [
     'libemmalloc-memvalidate',
     'libemmalloc-verbose',
     'libemmalloc-memvalidate-verbose',
-    'libgl',
+    'libGL',
     'libhtml5',
     'libsockets',
     'libc_rt_wasm',
+    'libc_rt_wasm-optz',
     'struct_info',
     'libstandalonewasm',
-    'crt1'
+    'crt1',
+    'libunwind-except'
 ]
 
 USER_TASKS = [
@@ -87,6 +91,9 @@ USER_TASKS = [
 temp_files = shared.configuration.get_temp_files()
 logger = logging.getLogger('embuilder')
 force = False
+legacy_prefixes = {
+  'libgl': 'libGL',
+}
 
 
 def get_help():
@@ -101,15 +108,18 @@ Issuing 'embuilder.py build ALL' causes each task to be built.
 ''' % '\n        '.join(all_tasks)
 
 
-def build_port(port_name, lib_name):
+def build_port(port_name):
   if force:
-    shared.Cache.erase_file(shared.Cache.get_lib_name(lib_name))
+    system_libs.clear_port(port_name, settings)
 
-  system_libs.build_port(port_name, shared.Settings)
+  system_libs.build_port(port_name, settings)
 
 
 def main():
   global force
+
+  all_build_start_time = time.time()
+
   parser = argparse.ArgumentParser(description=__doc__,
                                    formatter_class=argparse.RawDescriptionHelpFormatter,
                                    epilog=get_help())
@@ -133,10 +143,10 @@ def main():
   shared.check_sanity()
 
   if args.lto:
-    shared.Settings.LTO = "full"
+    settings.LTO = "full"
 
   if args.pic:
-    shared.Settings.RELOCATABLE = 1
+    settings.RELOCATABLE = 1
 
   if args.force:
     force = True
@@ -163,7 +173,11 @@ def main():
     tasks = [x for x in tasks if x not in skip_tasks]
     print('Building targets: %s' % ' '.join(tasks))
   for what in tasks:
+    for old, new in legacy_prefixes.items():
+      if what.startswith(old):
+        what = what.replace(old, new)
     logger.info('building and verifying ' + what)
+    start_time = time.time()
     if what in SYSTEM_LIBRARIES:
       library = SYSTEM_LIBRARIES[what]
       if force:
@@ -178,87 +192,93 @@ def main():
         shared.Cache.erase_file('generated_struct_info.json')
       emscripten.generate_struct_info()
     elif what == 'icu':
-      build_port('icu', 'libicuuc.a')
+      build_port('icu')
     elif what == 'zlib':
-      shared.Settings.USE_ZLIB = 1
-      build_port('zlib', 'libz.a')
-      shared.Settings.USE_ZLIB = 0
+      settings.USE_ZLIB = 1
+      build_port('zlib')
+      settings.USE_ZLIB = 0
     elif what == 'bzip2':
-      build_port('bzip2', 'libbz2.a')
+      build_port('bzip2')
     elif what == 'bullet':
-      build_port('bullet', 'libbullet.a')
+      build_port('bullet')
     elif what == 'vorbis':
-      build_port('vorbis', 'libvorbis.a')
+      build_port('vorbis')
     elif what == 'ogg':
-      build_port('ogg', 'libogg.a')
+      build_port('ogg')
     elif what == 'giflib':
-      build_port('giflib', 'libgif.a')
+      build_port('giflib')
     elif what == 'libjpeg':
-      build_port('libjpeg', 'libjpeg.a')
+      build_port('libjpeg')
     elif what == 'libpng':
-      build_port('libpng', 'libpng.a')
+      build_port('libpng')
     elif what == 'sdl2':
-      build_port('sdl2', 'libSDL2.a')
+      build_port('sdl2')
     elif what == 'sdl2-mt':
-      shared.Settings.USE_PTHREADS = 1
-      build_port('sdl2', 'libSDL2-mt.a')
-      shared.Settings.USE_PTHREADS = 0
+      settings.USE_PTHREADS = 1
+      build_port('sdl2')
+      settings.USE_PTHREADS = 0
     elif what == 'sdl2-gfx':
-      build_port('sdl2_gfx', 'libSDL2_gfx.a')
+      build_port('sdl2_gfx')
     elif what == 'sdl2-image':
-      build_port('sdl2_image', 'libSDL2_image.a')
+      build_port('sdl2_image')
     elif what == 'sdl2-image-png':
-      shared.Settings.SDL2_IMAGE_FORMATS = ["png"]
-      build_port('sdl2_image', 'libSDL2_image_png.a')
-      shared.Settings.SDL2_IMAGE_FORMATS = []
+      settings.SDL2_IMAGE_FORMATS = ["png"]
+      build_port('sdl2_image')
+      settings.SDL2_IMAGE_FORMATS = []
     elif what == 'sdl2-image-jpg':
-      shared.Settings.SDL2_IMAGE_FORMATS = ["jpg"]
-      build_port('sdl2_image', 'libSDL2_image_jpg.a')
-      shared.Settings.SDL2_IMAGE_FORMATS = []
+      settings.SDL2_IMAGE_FORMATS = ["jpg"]
+      build_port('sdl2_image')
+      settings.SDL2_IMAGE_FORMATS = []
     elif what == 'sdl2-net':
-      build_port('sdl2_net', 'libSDL2_net.a')
+      build_port('sdl2_net')
     elif what == 'sdl2-mixer':
-      old_formats = shared.Settings.SDL2_MIXER_FORMATS
-      shared.Settings.SDL2_MIXER_FORMATS = []
-      build_port('sdl2_mixer', 'libSDL2_mixer.a')
-      shared.Settings.SDL2_MIXER_FORMATS = old_formats
+      old_formats = settings.SDL2_MIXER_FORMATS
+      settings.SDL2_MIXER_FORMATS = []
+      build_port('sdl2_mixer')
+      settings.SDL2_MIXER_FORMATS = old_formats
     elif what == 'sdl2-mixer-ogg':
-      old_formats = shared.Settings.SDL2_MIXER_FORMATS
-      shared.Settings.SDL2_MIXER_FORMATS = ["ogg"]
-      build_port('sdl2_mixer', 'libSDL2_mixer_ogg.a')
-      shared.Settings.SDL2_MIXER_FORMATS = old_formats
+      old_formats = settings.SDL2_MIXER_FORMATS
+      settings.SDL2_MIXER_FORMATS = ["ogg"]
+      build_port('sdl2_mixer')
+      settings.SDL2_MIXER_FORMATS = old_formats
     elif what == 'sdl2-mixer-mp3':
-      old_formats = shared.Settings.SDL2_MIXER_FORMATS
-      shared.Settings.SDL2_MIXER_FORMATS = ["mp3"]
-      build_port('sdl2_mixer', 'libSDL2_mixer_mp3.a')
-      shared.Settings.SDL2_MIXER_FORMATS = old_formats
+      old_formats = settings.SDL2_MIXER_FORMATS
+      settings.SDL2_MIXER_FORMATS = ["mp3"]
+      build_port('sdl2_mixer')
+      settings.SDL2_MIXER_FORMATS = old_formats
     elif what == 'freetype':
-      build_port('freetype', 'libfreetype.a')
+      build_port('freetype')
     elif what == 'harfbuzz':
-      build_port('harfbuzz', 'libharfbuzz.a')
+      build_port('harfbuzz')
     elif what == 'harfbuzz-mt':
-      shared.Settings.USE_PTHREADS = 1
-      build_port('harfbuzz', 'libharfbuzz-mt.a')
-      shared.Settings.USE_PTHREADS = 0
+      settings.USE_PTHREADS = 1
+      build_port('harfbuzz')
+      settings.USE_PTHREADS = 0
     elif what == 'sdl2-ttf':
-      build_port('sdl2_ttf', 'libSDL2_ttf.a')
+      build_port('sdl2_ttf')
     elif what == 'cocos2d':
-      build_port('cocos2d', 'libcocos2d.a')
+      build_port('cocos2d')
     elif what == 'regal':
-      build_port('regal', 'libregal.a')
+      build_port('regal')
     elif what == 'regal-mt':
-      shared.Settings.USE_PTHREADS = 1
-      build_port('regal', 'libregal-mt.a')
-      shared.Settings.USE_PTHREADS = 0
+      settings.USE_PTHREADS = 1
+      build_port('regal')
+      settings.USE_PTHREADS = 0
     elif what == 'boost_headers':
-      build_port('boost_headers', 'libboost_headers.a')
+      build_port('boost_headers')
     elif what == 'mpg123':
-      build_port('mpg123', 'libmpg123.a')
+      build_port('mpg123')
     else:
       logger.error('unfamiliar build target: ' + what)
       return 1
 
-    logger.info('...success')
+    time_taken = time.time() - start_time
+    logger.info('...success. Took %s(%.2fs)' % (('%02d:%02d mins ' % (time_taken // 60, time_taken % 60) if time_taken >= 60 else ''), time_taken))
+
+  if len(tasks) > 1:
+    all_build_time_taken = time.time() - all_build_start_time
+    logger.info('Built %d targets in %s(%.2fs)' % (len(tasks), ('%02d:%02d mins ' % (all_build_time_taken // 60, all_build_time_taken % 60) if all_build_time_taken >= 60 else ''), all_build_time_taken))
+
   return 0
 
 
