@@ -5,39 +5,30 @@
 
 var LibraryDylink = {
 #if RELOCATABLE
-  $asmjsMangle: function(x) {
-    var unmangledSymbols = {{{ buildStringArray(WASM_SYSTEM_EXPORTS) }}};
-    return x.indexOf('dynCall_') == 0 || unmangledSymbols.indexOf(x) != -1 ? x : '_' + x;
-  },
-
   $resolveGlobalSymbol__deps: ['$asmjsMangle'],
   $resolveGlobalSymbol: function(symName, direct) {
     var sym;
 #if !WASM_BIGINT
     if (direct) {
       // First look for the orig$ symbol which is the symbols without
-      // any legalization performed.   Here we look on the 'asm' object
-      // to avoid any JS wrapping of the symbol.
-      sym = Module['asm']['orig$' + symName];
+      // any legalization performed.
+      sym = asmLibraryArg['orig$' + symName];
     }
 #endif
-    // Then look for the unmangled name itself.
     if (!sym) {
-      sym = Module['asm'][symName];
+      sym = asmLibraryArg[symName];
     }
-    // fall back to the mangled name on the module object which could include
-    // JavaScript functions and wrapped native functions.
-#if !WASM_BIGINT
-    if (!sym && direct) {
-      sym = Module['_orig$' + symName];
-    }
-#endif
 
+    // Check for the symbol on the Module object.  This is the only
+    // way to dynamically access JS library symbols that were not
+    // referenced by the main module (and therefore not part of the
+    // initial set of symbols included in asmLibraryArg when it
+    // was declared.
     if (!sym) {
       sym = Module[asmjsMangle(symName)];
     }
 
-    if (!sym && symName.indexOf('invoke_') == 0) {
+    if (!sym && symName.startsWith('invoke_')) {
       sym = createInvokeFunction(symName.split('_')[1]);
     }
 
@@ -52,7 +43,7 @@ var LibraryDylink = {
   $GOTHandler: {
     'get': function(obj, symName) {
       if (!GOT[symName]) {
-        GOT[symName] = new WebAssembly.Global({value: 'i32', mutable: true});
+        GOT[symName] = new WebAssembly.Global({'value': 'i32', 'mutable': true});
 #if DYLINK_DEBUG
         err("new GOT entry: " + symName);
 #endif
@@ -68,7 +59,7 @@ var LibraryDylink = {
       '__wasm_apply_data_relocs',
       '__dso_handle',
       '__set_stack_limits'
-    ].indexOf(symName) != -1
+    ].includes(symName)
 #if SPLIT_MODULE
         // Exports synthesized by wasm-split should be prefixed with '%'
         || symName[0] == '%'
@@ -89,14 +80,14 @@ var LibraryDylink = {
       var replace = false;
       var value = exports[symName];
 #if !WASM_BIGINT
-      if (symName.indexOf('orig$') == 0) {
+      if (symName.startsWith('orig$')) {
         symName = symName.split('$')[1];
         replace = true;
       }
 #endif
 
       if (!GOT[symName]) {
-        GOT[symName] = new WebAssembly.Global({value: 'i32', mutable: true});
+        GOT[symName] = new WebAssembly.Global({'value': 'i32', 'mutable': true});
       }
       if (replace || GOT[symName].value == 0) {
         if (typeof value === 'function') {
@@ -220,6 +211,7 @@ var LibraryDylink = {
   // Dynamic version of shared.py:make_invoke.  This is needed for invokes
   // that originate from side modules since these are not known at JS
   // generation time.
+  $createInvokeFunction__deps: ['$dynCall', 'setThrew'],
   $createInvokeFunction: function(sig) {
     return function() {
       var sp = stackSave();
@@ -239,8 +231,8 @@ var LibraryDylink = {
   // because the main program needs those linked in before it runs (so we can't
   // use normally malloc from the main program to do these allocations).
 
-  // Allocate memory no even if malloc isn't ready yet.
-  $getMemory__deps: ['$GOT'],
+  // Allocate memory even if malloc isn't ready yet.
+  $getMemory__deps: ['$GOT', '__heap_base'],
   $getMemory: function(size) {
     // After the runtime is initialized, we must only use sbrk() normally.
 #if DYLINK_DEBUG
@@ -248,12 +240,12 @@ var LibraryDylink = {
 #endif
     if (runtimeInitialized)
       return _malloc(size);
-    var ret = Module['___heap_base'];
+    var ret = ___heap_base;
     var end = (ret + size + 15) & -16;
 #if ASSERTIONS
     assert(end <= HEAP8.length, 'failure to getMemory - memory growth etc. is not supported there, call malloc/sbrk directly or increase INITIAL_MEMORY');
 #endif
-    Module['___heap_base'] = end;
+    ___heap_base = end;
     GOT['__heap_base'].value = end;
     return ret;
   },
@@ -286,27 +278,6 @@ var LibraryDylink = {
       return ret;
     }
 
-    function parseDylinkSection() {
-      var customSection = {};
-      customSection.memorySize = getLEB();
-      customSection.memoryAlign = getLEB();
-      customSection.tableSize = getLEB();
-      customSection.tableAlign = getLEB();
-      // shared libraries this module needs. We need to load them first, so that
-      // current module could resolve its imports. (see tools/shared.py
-      // WebAssembly.make_shared_library() for "dylink" section extension format)
-      var neededDynlibsCount = getLEB();
-      customSection.neededDynlibs = [];
-      for (var i = 0; i < neededDynlibsCount; ++i) {
-        var nameLen = getLEB();
-        var nameUTF8 = binary.subarray(next, next + nameLen);
-        next += nameLen;
-        var name = UTF8ArrayToString(nameUTF8, 0);
-        customSection.neededDynlibs.push(name);
-      }
-      return customSection;
-    }
-
     if (binary instanceof WebAssembly.Module) {
       var dylinkSection = WebAssembly.Module.customSections(binary, "dylink");
       assert(dylinkSection.length != 0, 'need dylink section');
@@ -327,15 +298,36 @@ var LibraryDylink = {
       assert(binary[next] === 'k'.charCodeAt(0)); next++;
     }
 
-    return parseDylinkSection();
+    var customSection = {};
+    customSection.memorySize = getLEB();
+    customSection.memoryAlign = getLEB();
+    customSection.tableSize = getLEB();
+    customSection.tableAlign = getLEB();
+#if ASSERTIONS
+    var tableAlign = Math.pow(2, customSection.tableAlign);
+    assert(tableAlign === 1, 'invalid tableAlign ' + tableAlign);
+#endif
+    // shared libraries this module needs. We need to load them first, so that
+    // current module could resolve its imports. (see tools/shared.py
+    // WebAssembly.make_shared_library() for "dylink" section extension format)
+    var neededDynlibsCount = getLEB();
+    customSection.neededDynlibs = [];
+    for (var i = 0; i < neededDynlibsCount; ++i) {
+      var nameLen = getLEB();
+      var nameUTF8 = binary.subarray(next, next + nameLen);
+      next += nameLen;
+      var name = UTF8ArrayToString(nameUTF8, 0);
+      customSection.neededDynlibs.push(name);
+    }
+    return customSection;
   },
 
   // Module.symbols <- libModule.symbols (flags.global handler)
   $mergeLibSymbols__deps: ['$asmjsMangle'],
-  $mergeLibSymbols: function(libModule, libName) {
+  $mergeLibSymbols: function(exports, libName) {
     // add symbols into global namespace TODO: weak linking etc.
-    for (var sym in libModule) {
-      if (!libModule.hasOwnProperty(sym)) {
+    for (var sym in exports) {
+      if (!exports.hasOwnProperty(sym)) {
         continue;
       }
 
@@ -344,20 +336,25 @@ var LibraryDylink = {
       //
       // We should copy the symbols (which include methods and variables) from SIDE_MODULE to MAIN_MODULE.
 
-      var module_sym = asmjsMangle(sym);
-
-      if (!Module.hasOwnProperty(module_sym)) {
-        Module[module_sym] = libModule[sym];
+      if (!asmLibraryArg.hasOwnProperty(sym)) {
+        asmLibraryArg[sym] = exports[sym];
       }
 #if ASSERTIONS == 2
       else {
-        var curr = Module[sym], next = libModule[sym];
+        var curr = asmLibraryArg[sym], next = exports[sym];
         // don't warn on functions - might be odr, linkonce_odr, etc.
         if (!(typeof curr === 'function' && typeof next === 'function')) {
           err("warning: symbol '" + sym + "' from '" + libName + "' already exists (duplicate symbol? or weak linking, which isn't supported yet?)"); // + [curr, ' vs ', next]);
         }
       }
 #endif
+
+      // Export native export on the Module object.
+      // TODO(sbc): Do all users want this?  Should we skip this by default?
+      var module_sym = asmjsMangle(sym);
+      if (!Module.hasOwnProperty(module_sym)) {
+        Module[module_sym] = exports[sym];
+      }
     }
   },
 
@@ -366,47 +363,43 @@ var LibraryDylink = {
   $loadWebAssemblyModule__deps: ['$loadDynamicLibrary', '$createInvokeFunction', '$getMemory', '$relocateExports', '$resolveGlobalSymbol', '$GOTHandler', '$getDylinkMetadata'],
   $loadWebAssemblyModule: function(binary, flags) {
     var metadata = getDylinkMetadata(binary);
-    var memorySize = metadata.memorySize;
-    var memoryAlign = metadata.memoryAlign;
-    var tableSize = metadata.tableSize;
-    var tableAlign = metadata.tableAlign;
-    var neededDynlibs = metadata.neededDynlibs;
+#if ASSERTIONS
+    var originalTable = wasmTable;
+#endif
 
     // loadModule loads the wasm module after all its dependencies have been loaded.
     // can be called both sync/async.
     function loadModule() {
       // alignments are powers of 2
-      memoryAlign = Math.pow(2, memoryAlign);
-      tableAlign = Math.pow(2, tableAlign);
+      var memAlign = Math.pow(2, metadata.memoryAlign);
       // finalize alignments and verify them
-      memoryAlign = Math.max(memoryAlign, STACK_ALIGN); // we at least need stack alignment
-#if ASSERTIONS
-      assert(tableAlign === 1, 'invalid tableAlign ' + tableAlign);
-#endif
+      memAlign = Math.max(memAlign, STACK_ALIGN); // we at least need stack alignment
       // prepare memory
-      var memoryBase = alignMemory(getMemory(memorySize + memoryAlign), memoryAlign); // TODO: add to cleanups
+      var memoryBase = alignMemory(getMemory(metadata.memorySize + memAlign), memAlign); // TODO: add to cleanups
 #if DYLINK_DEBUG
       err("loadModule: memoryBase=" + memoryBase);
 #endif
-      // prepare env imports
-      var env = asmLibraryArg;
       // TODO: use only __memory_base and __table_base, need to update asm.js backend
-      var table = wasmTable;
-      var tableBase = table.length;
-      var originalTable = table;
-      table.grow(tableSize);
-      assert(table === originalTable);
+      var tableBase = wasmTable.length;
+      wasmTable.grow(metadata.tableSize);
       // zero-initialize memory and table
       // The static area consists of explicitly initialized data, followed by zero-initialized data.
       // The latter may need zeroing out if the MAIN_MODULE has already used this memory area before
       // dlopen'ing the SIDE_MODULE.  Since we don't know the size of the explicitly initialized data
       // here, we just zero the whole thing, which is suboptimal, but should at least resolve bugs
       // from uninitialized memory.
-      for (var i = memoryBase; i < memoryBase + memorySize; i++) {
-        HEAP8[i] = 0;
+#if USE_PTHREADS
+      // in a pthread the module heap was already allocated and initialized in the main thread.
+      if (!ENVIRONMENT_IS_PTHREAD) {
+#endif
+        for (var i = memoryBase; i < memoryBase + metadata.memorySize; i++) {
+          HEAP8[i] = 0;
+        }
+#if USE_PTHREADS
       }
-      for (var i = tableBase; i < tableBase + tableSize; i++) {
-        table.set(i, null);
+#endif
+      for (var i = tableBase; i < tableBase + metadata.tableSize; i++) {
+        wasmTable.set(i, null);
       }
 
       // This is the export map that we ultimately return.  We declare it here
@@ -428,13 +421,6 @@ var LibraryDylink = {
         return resolved;
       }
 
-      // copy currently exported symbols so the new module can import them
-      for (var x in Module) {
-        if (!(x in env)) {
-          env[x] = Module[x];
-        }
-      }
-
       // TODO kill ↓↓↓ (except "symbols local to this module", it will likely be
       // not needed if we require that if A wants symbols from B it has to link
       // to B explicitly: similarly to -Wl,--no-undefined)
@@ -447,7 +433,7 @@ var LibraryDylink = {
       // are. To do that here, we use a JS proxy (another option would
       // be to inspect the binary directly).
       var proxyHandler = {
-        'get': function(obj, prop) {
+        'get': function(stubs, prop) {
           // symbols that should be local to this module
           switch (prop) {
             case '__memory_base':
@@ -455,21 +441,26 @@ var LibraryDylink = {
             case '__table_base':
               return tableBase;
           }
-          if (prop in obj) {
-            return obj[prop]; // already present
+          if (prop in asmLibraryArg) {
+            // No stub needed, symbol already exists in symbol table
+            return asmLibraryArg[prop];
           }
-          // otherwise this is regular function import - call it indirectly
-          var resolved;
-          return obj[prop] = function() {
-            if (!resolved) resolved = resolveSymbol(prop, true);
-            return resolved.apply(null, arguments);
-          };
+          // Return a stub function that will resolve the symbol
+          // when first called.
+          if (!(prop in stubs)) {
+            var resolved;
+            stubs[prop] = function() {
+              if (!resolved) resolved = resolveSymbol(prop, true);
+              return resolved.apply(null, arguments);
+            };
+          }
+          return stubs[prop];
         }
       };
-      var proxy = new Proxy(env, proxyHandler);
+      var proxy = new Proxy({}, proxyHandler);
       var info = {
-        'GOT.mem': new Proxy(asmLibraryArg, GOTHandler),
-        'GOT.func': new Proxy(asmLibraryArg, GOTHandler),
+        'GOT.mem': new Proxy({}, GOTHandler),
+        'GOT.func': new Proxy({}, GOTHandler),
         'env': proxy,
         {{{ WASI_MODULE_NAME }}}: proxy,
       };
@@ -477,12 +468,11 @@ var LibraryDylink = {
       function postInstantiation(instance) {
 #if ASSERTIONS
         // the table should be unchanged
-        assert(table === originalTable);
-        assert(table === wasmTable);
+        assert(wasmTable === originalTable);
 #endif
         // add new entries to functionsInTableMap
-        for (var i = 0; i < tableSize; i++) {
-          var item = table.get(tableBase + i);
+        for (var i = 0; i < metadata.tableSize; i++) {
+          var item = wasmTable.get(tableBase + i);
 #if ASSERTIONS
           // verify that the new table region was filled in
           assert(item !== undefined, 'table entry was not filled in');
@@ -499,16 +489,38 @@ var LibraryDylink = {
 #if STACK_OVERFLOW_CHECK >= 2
         moduleExports['__set_stack_limits']({{{ STACK_BASE }}} , {{{ STACK_MAX }}});
 #endif
+
         // initialize the module
-        var init = moduleExports['__post_instantiate'];
-        if (init) {
-          if (runtimeInitialized) {
-            init();
-          } else {
-            // we aren't ready to run compiled code yet
-            __ATINIT__.push(init);
+#if USE_PTHREADS
+        // The main thread does a full __wasm_call_ctors (which includes a call
+        // to emscripten_tls_init), but secondary threads should not call static
+        // constructors in general - emscripten_tls_init is the exception.
+        if (ENVIRONMENT_IS_PTHREAD) {
+          var init = moduleExports['emscripten_tls_init'];
+          assert(init);
+#if DYLINK_DEBUG
+          out("adding to tlsInitFunctions: " + init);
+#endif
+          PThread.tlsInitFunctions.push(init);
+        } else {
+#endif
+          var init = moduleExports['__wasm_call_ctors'];
+          // TODO(sbc): Remove this once extra check once the binaryen
+          // change propogates: https://github.com/WebAssembly/binaryen/pull/3811
+          if (!init) {
+            init = moduleExports['__post_instantiate'];
           }
+          if (init) {
+            if (runtimeInitialized) {
+              init();
+            } else {
+              // we aren't ready to run compiled code yet
+              __ATINIT__.push(init);
+            }
+          }
+#if USE_PTHREADS
         }
+#endif
         return moduleExports;
       }
 
@@ -529,7 +541,7 @@ var LibraryDylink = {
 
     // now load needed libraries and the module itself.
     if (flags.loadAsync) {
-      return neededDynlibs.reduce(function(chain, dynNeeded) {
+      return metadata.neededDynlibs.reduce(function(chain, dynNeeded) {
         return chain.then(function() {
           return loadDynamicLibrary(dynNeeded, flags);
         });
@@ -538,7 +550,7 @@ var LibraryDylink = {
       });
     }
 
-    neededDynlibs.forEach(function(dynNeeded) {
+    metadata.neededDynlibs.forEach(function(dynNeeded) {
       loadDynamicLibrary(dynNeeded, flags);
     });
     return loadModule();
@@ -674,11 +686,7 @@ var LibraryDylink = {
 #if DYLINK_DEBUG
     err('preloadDylibs');
 #endif
-    var libs = {{{ JSON.stringify(RUNTIME_LINKED_LIBS) }}};
-    if (Module['dynamicLibraries']) {
-      libs = libs.concat(Module['dynamicLibraries'])
-    }
-    if (!libs.length) {
+    if (!dynamicLibraries.length) {
 #if DYLINK_DEBUG
       err('preloadDylibs: no libraries to preload');
 #endif
@@ -690,7 +698,7 @@ var LibraryDylink = {
     if (!readBinary) {
       // we can't read binary data synchronously, so preload
       addRunDependency('preloadDylibs');
-      libs.reduce(function(chain, lib) {
+      dynamicLibraries.reduce(function(chain, lib) {
         return chain.then(function() {
           return loadDynamicLibrary(lib, {loadAsync: true, global: true, nodelete: true, allowUndefined: true});
         });
@@ -702,7 +710,7 @@ var LibraryDylink = {
       return;
     }
 
-    libs.forEach(function(lib) {
+    dynamicLibraries.forEach(function(lib) {
       // libraries linked to main never go away
       loadDynamicLibrary(lib, {global: true, nodelete: true, allowUndefined: true});
     });
