@@ -110,7 +110,7 @@ var LibraryExceptions = {
     };
   },
 
-  $CatchInfo__deps: ['$ExceptionInfo'],
+  $CatchInfo__deps: ['$ExceptionInfo', '__cxa_is_pointer_type'],
   // This native structure is returned from __cxa_find_matching_catch, and serves as catching
   // context, i.e. stores information required to proceed with a specific selected catch. It stores
   // base and adjusted pointers of a thrown object. It is allocated dynamically and should be freed
@@ -206,12 +206,14 @@ var LibraryExceptions = {
 
   // Exceptions
   __cxa_allocate_exception__deps: ['$ExceptionInfoAttrs'],
+  __cxa_allocate_exception__sig: 'vi',
   __cxa_allocate_exception: function(size) {
     // Thrown object is prepended by exception metadata block
     return _malloc(size + ExceptionInfoAttrs.SIZE) + ExceptionInfoAttrs.SIZE;
   },
 
   __cxa_free_exception__deps: ['$ExceptionInfo'],
+  __cxa_free_exception__sig: 'vi',
   __cxa_free_exception: function(ptr) {
 #if ABORTING_MALLOC || ASSERTIONS
     try {
@@ -261,6 +263,9 @@ var LibraryExceptions = {
   __cxa_rethrow__sig: 'v',
   __cxa_rethrow: function() {
     var catchInfo = exceptionCaught.pop();
+    if (!catchInfo) {
+      abort('no exception to throw');
+    }
     var info = catchInfo.get_exception_info();
     var ptr = catchInfo.get_base_ptr();
     if (!info.get_rethrown()) {
@@ -377,7 +382,7 @@ var LibraryExceptions = {
   // unwinding using 'if' blocks around each function, so the remaining
   // functionality boils down to picking a suitable 'catch' block.
   // We'll do that here, instead, to keep things simpler.
-  __cxa_find_matching_catch__deps: ['$exceptionLast', '$ExceptionInfo', '$CatchInfo', '__resumeException'],
+  __cxa_find_matching_catch__deps: ['$exceptionLast', '$ExceptionInfo', '$CatchInfo', '__resumeException', '__cxa_can_catch'],
   __cxa_find_matching_catch: function() {
     var thrown = exceptionLast;
     if (!thrown) {
@@ -426,8 +431,7 @@ var LibraryExceptions = {
     {{{ makeStructuralReturn(['catchInfo.ptr', 'thrownType']) }}};
   },
 
-  __resumeException__deps: [function() { '$exceptionLast', '$CatchInfo',
-                                        Functions.libraryFunctions['___resumeException'] = 1 }], // will be called directly from compiled code
+  __resumeException__deps: ['$exceptionLast', '$CatchInfo'],
   __resumeException: function(catchInfoPtr) {
     var catchInfo = new CatchInfo(catchInfoPtr);
     var ptr = catchInfo.get_base_ptr();
