@@ -4,10 +4,11 @@
 
 #include <stdint.h>
 #include <string.h>
+#include <emscripten/emscripten.h>
 
 // An external JS implementation that is efficient for very large copies, using
 // HEAPU8.set()
-extern void *emscripten_memcpy_big(void *restrict dest, const void *restrict src, size_t n);
+void* emscripten_memcpy_big(void *restrict dest, const void *restrict src, size_t n) EM_IMPORT(emscripten_memcpy_big);
 
 // XXX EMSCRIPTEN ASAN: build an uninstrumented version of memcpy
 #if defined(__EMSCRIPTEN__) && defined(__has_feature)
@@ -16,8 +17,19 @@ extern void *emscripten_memcpy_big(void *restrict dest, const void *restrict src
 #endif
 #endif
 
-void *memcpy(void *restrict dest, const void *restrict src, size_t n)
-{
+#ifdef EMSCRIPTEN_OPTIMIZE_FOR_OZ
+
+void *memcpy(void *dest, const void *src, size_t n) {
+  unsigned char *d = (unsigned char *)dest;
+  const unsigned char *s = (const unsigned char *)src;
+#pragma clang loop unroll(disable)
+  while(n--) *d++ = *s++;
+  return dest;
+}
+
+#else
+
+void *memcpy(void *restrict dest, const void *restrict src, size_t n) {
   unsigned char *d = dest;
   const unsigned char *s = src;
 
@@ -25,7 +37,7 @@ void *memcpy(void *restrict dest, const void *restrict src, size_t n)
   unsigned char *block_aligned_d_end;
   unsigned char *d_end;
 
-  if (n >= 8192) {
+  if (n >= 512) {
     emscripten_memcpy_big(dest, src, n);
     return dest;
   }
@@ -86,8 +98,6 @@ void *memcpy(void *restrict dest, const void *restrict src, size_t n)
     *d++ = *s++;
   }
   return dest;
-
-
-  for (; n; n--) *d++ = *s++;
-  return dest;
 }
+
+#endif
