@@ -5,6 +5,7 @@
 
 #include <assert.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <dlfcn.h>
 #include <emscripten.h>
 
@@ -16,7 +17,7 @@ voidfunc onefunc;
 intfunc twofunc;
 
 void next(const char *x) {
-  lib_handle = dlopen("themodule.js", RTLD_NOW);
+  lib_handle = dlopen("thelib.wasm", RTLD_NOW);
   assert(lib_handle != NULL);
 
   onefunc = (voidfunc)dlsym(lib_handle, "one");
@@ -36,12 +37,12 @@ void next(const char *x) {
   assert(twofunc() == 7);
   onefunc();
   int result = twofunc();
-  REPORT_RESULT(result);
+  exit(result);
 }
 
 int main() {
-  emscripten_async_wget("module.js", "themodule.js", next, NULL);
-  
+  emscripten_async_wget("lib.wasm", "thelib.wasm", next, NULL);
+  printf("returning from main\n");
   return 0;
 }
 
