@@ -9,7 +9,6 @@ import sys
 from subprocess import Popen, PIPE, CalledProcessError
 
 from tools import shared
-from tools.shared import config
 
 WORKING_ENGINES = {} # Holds all configured engines and whether they work: maps path -> True/False
 
@@ -63,7 +62,7 @@ def check_engine(engine):
   if engine_path not in WORKING_ENGINES:
     logging.debug('Checking JS engine %s' % engine)
     try:
-      output = run_js(shared.path_from_root('tests', 'hello_world.js'), engine, skip_check=True)
+      output = run_js(shared.path_from_root('tests/hello_world.js'), engine, skip_check=True)
       if 'hello, world!' in output:
         WORKING_ENGINES[engine_path] = True
       else:
@@ -87,12 +86,9 @@ def require_engine(engine):
     sys.exit(1)
 
 
-def run_js(filename, engine=None, args=[],
+def run_js(filename, engine, args=[],
            stdin=None, stdout=PIPE, stderr=None, cwd=None,
            full_output=False, assert_returncode=0, skip_check=False):
-  if not engine:
-    engine = config.JS_ENGINES[0]
-
   # We used to support True here but we no longer do.  Assert here just in case.
   assert(type(assert_returncode) == int)
 
