@@ -3,58 +3,62 @@
 # University of Illinois/NCSA Open Source License.  Both these licenses can be
 # found in the LICENSE file.
 
-import os, shutil, logging
-from subprocess import Popen
+import os
+import shutil
+
+TAG = '1.2.11'
+HASH = 'a42b8359e76cf7b3ae70bf31f0f8a8caa407ac80e8fe08b838076cd5e45ac2e685dae45eb59db2d25543fb3b5bd13b843a02bb8373cda704d7238be50d5e9c68'
+
+
+def needed(settings):
+  return settings.USE_ZLIB
 
-TAG = 'version_1'
 
 def get(ports, settings, shared):
-  if settings.USE_ZLIB == 1:
-    ports.fetch_project('zlib', 'https://github.com/emscripten-ports/zlib/archive/' + TAG + '.zip', 'zlib-' + TAG)
-    def create():     
-      ports.clear_project_build('zlib')
+  ports.fetch_project('zlib', 'https://github.com/madler/zlib/archive/v' + TAG + '.zip', 'zlib-' + TAG, sha512hash=HASH)
 
-      source_path = os.path.join(ports.get_dir(), 'zlib', 'zlib-' + TAG)
-      dest_path = os.path.join(shared.Cache.get_path('ports-builds'), 'zlib')
-      shared.try_delete(dest_path)
-      os.makedirs(dest_path)
-      shutil.rmtree(dest_path, ignore_errors=True)
-      shutil.copytree(source_path, dest_path)
-      open(os.path.join(dest_path, 'zconf.h'), 'w').write(zconf_h)
+  def create(final):
+    ports.clear_project_build('zlib')
 
-      # build
-      srcs = 'adler32.c compress.c crc32.c deflate.c gzclose.c gzlib.c gzread.c gzwrite.c infback.c inffast.c inflate.c inftrees.c trees.c uncompr.c zutil.c'.split(' ')
-      commands = []
-      o_s = []
-      for src in srcs:
-        o = os.path.join(ports.get_build_dir(), 'zlib', src + '.o')
-        shared.safe_ensure_dirs(os.path.dirname(o))
-        commands.append([shared.PYTHON, shared.EMCC, os.path.join(dest_path, src), '-O2', '-o', o, '-I' + dest_path,'-w',])
-        o_s.append(o)
+    source_path = os.path.join(ports.get_dir(), 'zlib', 'zlib-' + TAG)
+    dest_path = os.path.join(ports.get_build_dir(), 'zlib')
+    shared.try_delete(dest_path)
+    os.makedirs(dest_path)
+    shutil.rmtree(dest_path, ignore_errors=True)
+    shutil.copytree(source_path, dest_path)
+    open(os.path.join(dest_path, 'zconf.h'), 'w').write(zconf_h)
+    ports.install_headers(dest_path)
 
-      ports.run_commands(commands)
-      final = os.path.join(ports.get_build_dir(), 'zlib', 'libz.a')
-      shared.try_delete(final)
-      Popen([shared.LLVM_AR, 'rc', final] + o_s).communicate()
-      assert os.path.exists(final)
-      return final
-    return [shared.Cache.get('zlib', create, what='port')]
-  else:
-    return []
+    # build
+    srcs = 'adler32.c compress.c crc32.c deflate.c gzclose.c gzlib.c gzread.c gzwrite.c infback.c inffast.c inflate.c inftrees.c trees.c uncompr.c zutil.c'.split()
+    commands = []
+    o_s = []
+    for src in srcs:
+      o = os.path.join(ports.get_build_dir(), 'zlib', src + '.o')
+      shared.safe_ensure_dirs(os.path.dirname(o))
+      commands.append([shared.EMCC, os.path.join(dest_path, src), '-O2', '-o', o, '-I' + dest_path, '-w', '-c'])
+      o_s.append(o)
+    ports.run_commands(commands)
+
+    ports.create_lib(final, o_s)
+
+  return [shared.Cache.get_lib('libz.a', create, what='port')]
+
+
+def clear(ports, settings, shared):
+  shared.Cache.erase_lib('libz.a')
+
+
+def process_args(ports):
+  return []
 
-def process_args(ports, args, settings, shared):
-  if settings.USE_ZLIB == 1:
-    get(ports, settings, shared)
-    args += ['-Xclang', '-isystem' + os.path.join(shared.Cache.get_path('ports-builds'), 'zlib')]
-  return args
 
 def show():
   return 'zlib (USE_ZLIB=1; zlib license)'
 
 
-
 zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
- * Copyright (C) 1995-2013 Jean-loup Gailly.
+ * Copyright (C) 1995-2016 Jean-loup Gailly, Mark Adler
  * For conditions of distribution and use, see copyright notice in zlib.h
  */
 
@@ -72,7 +76,7 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
 #ifdef Z_PREFIX     /* may be set to #if 1 by ./configure */
 #  define Z_PREFIX_SET
 
-/* all linked symbols */
+/* all linked symbols and init macros */
 #  define _dist_code            z__dist_code
 #  define _length_code          z__length_code
 #  define _tr_align             z__tr_align
@@ -84,6 +88,7 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
 #  define adler32               z_adler32
 #  define adler32_combine       z_adler32_combine
 #  define adler32_combine64     z_adler32_combine64
+#  define adler32_z             z_adler32_z
 #  ifndef Z_SOLO
 #    define compress              z_compress
 #    define compress2             z_compress2
@@ -92,10 +97,14 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
 #  define crc32                 z_crc32
 #  define crc32_combine         z_crc32_combine
 #  define crc32_combine64       z_crc32_combine64
+#  define crc32_z               z_crc32_z
 #  define deflate               z_deflate
 #  define deflateBound          z_deflateBound
 #  define deflateCopy           z_deflateCopy
 #  define deflateEnd            z_deflateEnd
+#  define deflateGetDictionary  z_deflateGetDictionary
+#  define deflateInit           z_deflateInit
+#  define deflateInit2          z_deflateInit2
 #  define deflateInit2_         z_deflateInit2_
 #  define deflateInit_          z_deflateInit_
 #  define deflateParams         z_deflateParams
@@ -122,6 +131,8 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
 #    define gzeof                 z_gzeof
 #    define gzerror               z_gzerror
 #    define gzflush               z_gzflush
+#    define gzfread               z_gzfread
+#    define gzfwrite              z_gzfwrite
 #    define gzgetc                z_gzgetc
 #    define gzgetc_               z_gzgetc_
 #    define gzgets                z_gzgets
@@ -133,7 +144,6 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
 #      define gzopen_w              z_gzopen_w
 #    endif
 #    define gzprintf              z_gzprintf
-#    define gzvprintf             z_gzvprintf
 #    define gzputc                z_gzputc
 #    define gzputs                z_gzputs
 #    define gzread                z_gzread
@@ -144,32 +154,39 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
 #    define gztell                z_gztell
 #    define gztell64              z_gztell64
 #    define gzungetc              z_gzungetc
+#    define gzvprintf             z_gzvprintf
 #    define gzwrite               z_gzwrite
 #  endif
 #  define inflate               z_inflate
 #  define inflateBack           z_inflateBack
 #  define inflateBackEnd        z_inflateBackEnd
+#  define inflateBackInit       z_inflateBackInit
 #  define inflateBackInit_      z_inflateBackInit_
+#  define inflateCodesUsed      z_inflateCodesUsed
 #  define inflateCopy           z_inflateCopy
 #  define inflateEnd            z_inflateEnd
+#  define inflateGetDictionary  z_inflateGetDictionary
 #  define inflateGetHeader      z_inflateGetHeader
+#  define inflateInit           z_inflateInit
+#  define inflateInit2          z_inflateInit2
 #  define inflateInit2_         z_inflateInit2_
 #  define inflateInit_          z_inflateInit_
 #  define inflateMark           z_inflateMark
 #  define inflatePrime          z_inflatePrime
 #  define inflateReset          z_inflateReset
 #  define inflateReset2         z_inflateReset2
+#  define inflateResetKeep      z_inflateResetKeep
 #  define inflateSetDictionary  z_inflateSetDictionary
-#  define inflateGetDictionary  z_inflateGetDictionary
 #  define inflateSync           z_inflateSync
 #  define inflateSyncPoint      z_inflateSyncPoint
 #  define inflateUndermine      z_inflateUndermine
-#  define inflateResetKeep      z_inflateResetKeep
+#  define inflateValidate       z_inflateValidate
 #  define inflate_copyright     z_inflate_copyright
 #  define inflate_fast          z_inflate_fast
 #  define inflate_table         z_inflate_table
 #  ifndef Z_SOLO
 #    define uncompress            z_uncompress
+#    define uncompress2           z_uncompress2
 #  endif
 #  define zError                z_zError
 #  ifndef Z_SOLO
@@ -279,9 +296,19 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
 #  define z_const
 #endif
 
-/* Some Mac compilers merge all .h files incorrectly: */
-#if defined(__MWERKS__)||defined(applec)||defined(THINK_C)||defined(__SC__)
-#  define NO_DUMMY_DECL
+#ifdef Z_SOLO
+   typedef unsigned long z_size_t;
+#else
+#  define z_longlong long long
+#  if defined(NO_SIZE_T)
+     typedef unsigned NO_SIZE_T z_size_t;
+#  elif defined(STDC)
+#    include <stddef.h>
+     typedef size_t z_size_t;
+#  else
+     typedef unsigned long z_size_t;
+#  endif
+#  undef z_longlong
 #endif
 
 /* Maximum value for memLevel in deflateInit2 */
@@ -311,7 +338,7 @@ zconf_h = r'''/* zconf.h -- configuration of the zlib compression library
  Of course this will generally degrade compression (there's no free lunch).
 
    The memory requirements for inflate are (in bytes) 1 << windowBits
- that is, 32K for windowBits=15 (default value) plus a few kilobytes
+ that is, 32K for windowBits=15 (default value) plus about 7 kilobytes
  for small objects.
 */
 
