@@ -1,13 +1,19 @@
+# Copyright 2014 The Emscripten Authors.  All rights reserved.
+# Emscripten is available under two separate licenses, the MIT license and the
+# University of Illinois/NCSA Open Source License.  Both these licenses can be
+# found in the LICENSE file.
 
-'''
-WebIDL binder
+"""WebIDL binder
 
-http://kripken.github.io/emscripten-site/docs/porting/connecting_cpp_and_javascript/WebIDL-Binder.html
-'''
+https://emscripten.org/docs/porting/connecting_cpp_and_javascript/WebIDL-Binder.html
+"""
 
-import os, sys
+import os
+import sys
 
-import shared
+sys.path.insert(1, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
+
+from tools import shared
 
 sys.path.append(shared.path_from_root('third_party'))
 sys.path.append(shared.path_from_root('third_party', 'ply'))
@@ -19,20 +25,26 @@ import WebIDL
 # CHECKS='ALL' will do extensive argument type checking (~5x slower than default).
 #                 This will catch invalid numbers, invalid pointers, invalid strings, etc.
 # Anything else defaults to legacy mode for backward compatibility.
-CHECKS = os.environ.get('IDL_CHECKS') or 'DEFAULT'
+CHECKS = os.environ.get('IDL_CHECKS', 'DEFAULT')
 # DEBUG=1 will print debug info in render_function
-DEBUG = os.environ.get('IDL_VERBOSE') is '1'
+DEBUG = os.environ.get('IDL_VERBOSE') == '1'
 
-if DEBUG: print "Debug print ON, CHECKS=%s" % CHECKS
+if DEBUG:
+  print("Debug print ON, CHECKS=%s" % CHECKS)
 
-class Dummy(object):
+# We need to avoid some closure errors on the constructors we define here.
+CONSTRUCTOR_CLOSURE_SUPPRESSIONS = '/** @suppress {undefinedVars, duplicate} @this{Object} */'
+
+
+class Dummy:
   def __init__(self, init):
-    for k, v in init.iteritems():
+    for k, v in init.items():
       self.__dict__[k] = v
 
-  def getExtendedAttribute(self, name):
+  def getExtendedAttribute(self, name): # noqa: U100
     return None
 
+
 input_file = sys.argv[1]
 output_base = sys.argv[2]
 
@@ -58,8 +70,8 @@ for thing in data:
   elif isinstance(thing, WebIDL.IDLEnum):
     enums[thing.identifier.name] = thing
 
-#print interfaces
-#print implements
+# print interfaces
+# print implements
 
 pre_c = []
 mid_c = []
@@ -73,31 +85,37 @@ mid_c += [r'''
 extern "C" {
 ''']
 
-def emit_constructor(name):
-  global mid_js
-  mid_js += [r'''%s.prototype = %s;
-%s.prototype.constructor = %s;
-%s.prototype.__class__ = %s;
-%s.__cache__ = {};
-Module['%s'] = %s;
-''' % (name, 'Object.create(%s.prototype)' % (implements[name][0] if implements.get(name) else 'WrapperObject'), name, name, name, name, name, name, name)]
+
+def build_constructor(name):
+  implementing_name = implements[name][0] if implements.get(name) else 'WrapperObject'
+  return [r'''{name}.prototype = Object.create({implementing}.prototype);
+{name}.prototype.constructor = {name};
+{name}.prototype.__class__ = {name};
+{name}.__cache__ = {{}};
+Module['{name}'] = {name};
+'''.format(name=name, implementing=implementing_name)]
 
 
 mid_js += ['''
 // Bindings utilities
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function WrapperObject() {
 }
 ''']
 
-emit_constructor('WrapperObject')
+mid_js += build_constructor('WrapperObject')
 
 mid_js += ['''
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant)
+    @param {*=} __class__ */
 function getCache(__class__) {
   return (__class__ || WrapperObject).__cache__;
 }
 Module['getCache'] = getCache;
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant)
+    @param {*=} __class__ */
 function wrapPointer(ptr, __class__) {
   var cache = getCache(__class__);
   var ret = cache[ptr];
@@ -108,6 +126,7 @@ function wrapPointer(ptr, __class__) {
 }
 Module['wrapPointer'] = wrapPointer;
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function castObject(obj, __class__) {
   return wrapPointer(obj.ptr, __class__);
 }
@@ -115,6 +134,7 @@ Module['castObject'] = castObject;
 
 Module['NULL'] = wrapPointer(0);
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function destroy(obj) {
   if (!obj['__destroy__']) throw 'Error: Cannot destroy object. (Did you create it yourself?)';
   obj['__destroy__']();
@@ -123,16 +143,19 @@ function destroy(obj) {
 }
 Module['destroy'] = destroy;
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function compare(obj1, obj2) {
   return obj1.ptr === obj2.ptr;
 }
 Module['compare'] = compare;
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function getPointer(obj) {
   return obj.ptr;
 }
 Module['getPointer'] = getPointer;
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function getClass(obj) {
   return obj.__class__;
 }
@@ -140,6 +163,7 @@ Module['getClass'] = getClass;
 
 // Converts big (string or array) values into a C-style storage, in temporary space
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 var ensureCache = {
   buffer: 0,  // the main buffer of temporary storage
   size: 0,   // the size of buffer
@@ -186,21 +210,22 @@ var ensureCache = {
       ensureCache.pos += len;
     }
     return ret;
-  },  
+  },
   copy: function(array, view, offset) {
-    var offsetShifted = offset;
+    offset >>>= 0;
     var bytes = view.BYTES_PER_ELEMENT;
     switch (bytes) {
-      case 2: offsetShifted >>= 1; break;
-      case 4: offsetShifted >>= 2; break;
-      case 8: offsetShifted >>= 3; break;
+      case 2: offset >>>= 1; break;
+      case 4: offset >>>= 2; break;
+      case 8: offset >>>= 3; break;
     }
     for (var i = 0; i < array.length; i++) {
-      view[offsetShifted + i] = array[i];
-    }   
+      view[offset + i] = array[i];
+    }
   },
 };
 
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function ensureString(value) {
   if (typeof value === 'string') {
     var intArray = intArrayFromString(value);
@@ -210,6 +235,7 @@ function ensureString(value) {
   }
   return value;
 }
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function ensureInt8(value) {
   if (typeof value === 'object') {
     var offset = ensureCache.alloc(value, HEAP8);
@@ -218,6 +244,7 @@ function ensureInt8(value) {
   }
   return value;
 }
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function ensureInt16(value) {
   if (typeof value === 'object') {
     var offset = ensureCache.alloc(value, HEAP16);
@@ -226,6 +253,7 @@ function ensureInt16(value) {
   }
   return value;
 }
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function ensureInt32(value) {
   if (typeof value === 'object') {
     var offset = ensureCache.alloc(value, HEAP32);
@@ -234,6 +262,7 @@ function ensureInt32(value) {
   }
   return value;
 }
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function ensureFloat32(value) {
   if (typeof value === 'object') {
     var offset = ensureCache.alloc(value, HEAPF32);
@@ -242,6 +271,7 @@ function ensureFloat32(value) {
   }
   return value;
 }
+/** @suppress {duplicate} (TODO: avoid emitting this multiple times, it is redundant) */
 function ensureFloat64(value) {
   if (typeof value === 'object') {
     var offset = ensureCache.alloc(value, HEAPF64);
@@ -255,117 +285,142 @@ function ensureFloat64(value) {
 
 mid_c += ['''
 // Not using size_t for array indices as the values used by the javascript code are signed.
+
+EM_JS(void, array_bounds_check_error, (size_t idx, size_t size), {
+  throw 'Array index ' + idx + ' out of bounds: [0,' + size + ')';
+});
+
 void array_bounds_check(const int array_size, const int array_idx) {
   if (array_idx < 0 || array_idx >= array_size) {
-    EM_ASM({
-      throw 'Array index ' + $0 + ' out of bounds: [0,' + $1 + ')';
-    }, array_idx, array_size);
+    array_bounds_check_error(array_idx, array_size);
   }
 }
 ''']
 
 C_FLOATS = ['float', 'double']
 
+
 def full_typename(arg):
-  return arg.type.name + ('[]' if arg.type.isArray() else '')
+  return ('const ' if arg.getExtendedAttribute('Const') else '') + arg.type.name + ('[]' if arg.type.isArray() else '')
+
 
 def type_to_c(t, non_pointing=False):
-  #print 'to c ', t
+  # print 'to c ', t
+  def base_type_to_c(t):
+    if t == 'Long':
+      ret = 'int'
+    elif t == 'UnsignedLong':
+      ret = 'unsigned int'
+    elif t == 'LongLong':
+      ret = 'long long'
+    elif t == 'UnsignedLongLong':
+      ret = 'unsigned long long'
+    elif t == 'Short':
+      ret = 'short'
+    elif t == 'UnsignedShort':
+      ret = 'unsigned short'
+    elif t == 'Byte':
+      ret = 'char'
+    elif t == 'Octet':
+      ret = 'unsigned char'
+    elif t == 'Void':
+      ret = 'void'
+    elif t == 'String':
+      ret = 'char*'
+    elif t == 'Float':
+      ret = 'float'
+    elif t == 'Double':
+      ret = 'double'
+    elif t == 'Boolean':
+      ret = 'bool'
+    elif t == 'Any' or t == 'VoidPtr':
+      ret = 'void*'
+    elif t in interfaces:
+      ret = (interfaces[t].getExtendedAttribute('Prefix') or [''])[0] + t + ('' if non_pointing else '*')
+    else:
+      ret = t
+    return ret
+
   t = t.replace(' (Wrapper)', '')
+
+  prefix = ''
   suffix = ''
   if '[]' in t:
-    suffix = '*'
     t = t.replace('[]', '')
-  if t == 'Long':
-    ret = 'int'
-  elif t == 'UnsignedLong':
-    ret = 'unsigned int'
-  elif t == 'Short':
-    ret = 'short'
-  elif t == 'UnsignedShort':
-    ret = 'unsigned short'
-  elif t == 'Byte':
-    ret = 'char'
-  elif t == 'Octet':
-    ret = 'unsigned char'
-  elif t == 'Void':
-    ret = 'void'
-  elif t == 'String':
-    ret = 'char*'
-  elif t == 'Float':
-    ret = 'float'
-  elif t == 'Double':
-    ret = 'double'
-  elif t == 'Boolean':
-    ret = 'bool'
-  elif t == 'Any' or t == 'VoidPtr':
-    ret = 'void*'
-  elif t in interfaces:
-    ret = (interfaces[t].getExtendedAttribute('Prefix') or [''])[0] + t + ('' if non_pointing else '*')
-  else:
-    ret = t
-  return ret + suffix
+    suffix = '*'
+  if 'const ' in t:
+    t = t.replace('const ', '')
+    prefix = 'const '
+  return prefix + base_type_to_c(t) + suffix
+
 
 def take_addr_if_nonpointer(m):
   if m.getExtendedAttribute('Ref') or m.getExtendedAttribute('Value'):
     return '&'
   return ''
 
+
 def deref_if_nonpointer(m):
   if m.getExtendedAttribute('Ref') or m.getExtendedAttribute('Value'):
     return '*'
   return ''
 
+
 def type_to_cdec(raw):
-  name = ret = type_to_c(raw.type.name, non_pointing=True)
-  if raw.getExtendedAttribute('Const'): ret = 'const ' + ret
-  if name not in interfaces: return ret
+  ret = type_to_c(raw.type.name, non_pointing=True)
+  if raw.getExtendedAttribute('Const'):
+    ret = 'const ' + ret
+  if raw.type.name not in interfaces:
+    return ret
   if raw.getExtendedAttribute('Ref'):
     return ret + '&'
   if raw.getExtendedAttribute('Value'):
     return ret
   return ret + '*'
 
-def render_function(class_name, func_name, sigs, return_type, non_pointer, copy, operator, constructor, func_scope, call_content=None, const=False, array_attribute=False):
+
+def render_function(class_name, func_name, sigs, return_type, non_pointer,
+                    copy, operator, constructor, func_scope,
+                    call_content=None, const=False, array_attribute=False):
   global mid_c, mid_js, js_impl_methods
 
   legacy_mode = CHECKS not in ['ALL', 'FAST']
-  all_checks = CHECKS  == 'ALL'
+  all_checks = CHECKS == 'ALL'
 
   bindings_name = class_name + '_' + func_name
   min_args = min(sigs.keys())
   max_args = max(sigs.keys())
 
-  c_names = {}
   all_args = sigs.get(max_args)
 
   if DEBUG:
-    print 'renderfunc', class_name, func_name, sigs.keys(), return_type, constructor
+    print('renderfunc', class_name, func_name, list(sigs.keys()), return_type, constructor)
     for i in range(max_args):
       a = all_args[i]
       if isinstance(a, WebIDL.IDLArgument):
-        print ("  arg%d" % i), a.identifier, a.type, a.optional
+        print(' ', a.identifier.name, a.identifier, a.type, a.optional)
       else:
-        print "  arg%d" % i
+        print("  arg%d" % i)
 
   # JS
 
   cache = ('getCache(%s)[this.ptr] = this;' % class_name) if constructor else ''
   call_prefix = '' if not constructor else 'this.ptr = '
   call_postfix = ''
-  if return_type != 'Void' and not constructor: call_prefix = 'return '
+  if return_type != 'Void' and not constructor:
+    call_prefix = 'return '
   if not constructor:
     if return_type in interfaces:
       call_prefix += 'wrapPointer('
       call_postfix += ', ' + return_type + ')'
     elif return_type == 'String':
-      call_prefix += 'Pointer_stringify('
+      call_prefix += 'UTF8ToString('
       call_postfix += ')'
     elif return_type == 'Boolean':
       call_prefix += '!!('
       call_postfix += ')'
 
-  args = ['arg%d' % i for i in range(max_args)]
+  args = [(all_args[i].identifier.name if isinstance(all_args[i], WebIDL.IDLArgument) else ('arg%d' % i)) for i in range(max_args)]
   if not constructor:
     body = '  var self = this.ptr;\n'
     pre_arg = ['self']
@@ -373,20 +428,16 @@ def render_function(class_name, func_name, sigs, return_type, non_pointer, copy,
     body = ''
     pre_arg = []
 
-  for i in range(max_args):
-    arg = all_args[i]
-    if arg.type.isString() or arg.type.isArray():
-      body += '  ensureCache.prepare();\n'
-      break
+  if any(arg.type.isString() or arg.type.isArray() for arg in all_args):
+    body += '  ensureCache.prepare();\n'
 
   full_name = "%s::%s" % (class_name, func_name)
 
-  for i in range(max_args):
+  for i, (js_arg, arg) in enumerate(zip(args, all_args)):
     if i >= min_args:
       optional = True
     else:
       optional = False
-    arg = all_args[i]
     do_default = False
     # Filter out arguments we don't know how to parse. Fast casing only common cases.
     compatible_arg = isinstance(arg, Dummy) or (isinstance(arg, WebIDL.IDLArgument) and arg.optional is False)
@@ -397,92 +448,101 @@ def render_function(class_name, func_name, sigs, return_type, non_pointer, copy,
       else:
         arg_name = ''
       # Format assert fail message
-      check_msg = "[CHECK FAILED] %s(arg%d:%s): " % (full_name, i, arg_name)
+      check_msg = "[CHECK FAILED] %s(%s:%s): " % (full_name, js_arg, arg_name)
       if isinstance(arg.type, WebIDL.IDLWrapperType):
         inner = arg.type.inner
       else:
         inner = ""
 
       # Print type info in comments.
-      body += "  /* arg%d <%s> [%s] */\n" % (i, arg.type.name, inner)
+      body += "  /* %s <%s> [%s] */\n" % (js_arg, arg.type.name, inner)
 
-      # Wrap asserts with existance check when argument is optional.
-      if all_checks and optional: body += "if(typeof arg%d !== 'undefined' && arg%d !== null) {\n" % (i, i)
+      # Wrap asserts with existence check when argument is optional.
+      if all_checks and optional:
+        body += "if(typeof {0} !== 'undefined' && {0} !== null) {{\n".format(js_arg)
       # Special case argument types.
       if arg.type.isNumeric():
         if arg.type.isInteger():
-          if all_checks: body += "  assert(typeof arg%d === 'number' && !isNaN(arg%d), '%sExpecting <integer>');\n" % (i, i, check_msg)
+          if all_checks:
+            body += "  assert(typeof {0} === 'number' && !isNaN({0}), '{1}Expecting <integer>');\n".format(js_arg, check_msg)
         else:
-          if all_checks: body += "  assert(typeof arg%d === 'number', '%sExpecting <number>');\n" % (i, check_msg)
+          if all_checks:
+            body += "  assert(typeof {0} === 'number', '{1}Expecting <number>');\n".format(js_arg, check_msg)
         # No transform needed for numbers
       elif arg.type.isBoolean():
-        if all_checks: body += "  assert(typeof arg%d === 'boolean' || (typeof arg%d === 'number' && !isNaN(arg%d)), '%sExpecting <boolean>');\n" % (i, i, i, check_msg)
+        if all_checks:
+          body += "  assert(typeof {0} === 'boolean' || (typeof {0} === 'number' && !isNaN({0})), '{1}Expecting <boolean>');\n".format(js_arg, check_msg)
         # No transform needed for booleans
       elif arg.type.isString():
         # Strings can be DOM strings or pointers.
-        if all_checks: body += "  assert(typeof arg%d === 'string' || (arg%d && typeof arg%d === 'object' && typeof arg%d.ptr === 'number'), '%sExpecting <string>');\n" % (i, i, i, i, check_msg)
+        if all_checks:
+          body += "  assert(typeof {0} === 'string' || ({0} && typeof {0} === 'object' && typeof {0}.ptr === 'number'), '{1}Expecting <string>');\n".format(js_arg, check_msg)
         do_default = True # legacy path is fast enough for strings.
       elif arg.type.isInterface():
-        if all_checks: body += "  assert(typeof arg%d === 'object' && typeof arg%d.ptr === 'number', '%sExpecting <pointer>');\n" % (i, i, check_msg)
+        if all_checks:
+          body += "  assert(typeof {0} === 'object' && typeof {0}.ptr === 'number', '{1}Expecting <pointer>');\n".format(js_arg, check_msg)
         if optional:
-          body += "  if(typeof arg%d !== 'undefined' && arg%d !== null) { arg%d = arg%d.ptr };\n" % (i, i, i, i)
+          body += "  if(typeof {0} !== 'undefined' && {0} !== null) {{ {0} = {0}.ptr }};\n".format(js_arg)
         else:
           # No checks in fast mode when the arg is required
-          body += "  arg%d = arg%d.ptr;\n" % (i, i)
+          body += "  {0} = {0}.ptr;\n".format(js_arg)
       else:
         do_default = True
 
-      if all_checks and optional: body += "}\n"
+      if all_checks and optional:
+        body += "}\n"
     else:
       do_default = True
 
     if do_default:
       if not (arg.type.isArray() and not array_attribute):
-        body += "  if (arg%d && typeof arg%d === 'object') arg%d = arg%d.ptr;\n" % (i, i, i, i)
+        body += "  if ({0} && typeof {0} === 'object') {0} = {0}.ptr;\n".format(js_arg)
         if arg.type.isString():
-          body += "  else arg%d = ensureString(arg%d);\n" % (i, i)
+          body += "  else {0} = ensureString({0});\n".format(js_arg)
       else:
         # an array can be received here
         arg_type = arg.type.name
         if arg_type in ['Byte', 'Octet']:
-          body += "  if (typeof arg%d == 'object') { arg%d = ensureInt8(arg%d); }\n" % (i, i, i)
+          body += "  if (typeof {0} == 'object') {{ {0} = ensureInt8({0}); }}\n".format(js_arg)
         elif arg_type in ['Short', 'UnsignedShort']:
-          body += "  if (typeof arg%d == 'object') { arg%d = ensureInt16(arg%d); }\n" % (i, i, i)
+          body += "  if (typeof {0} == 'object') {{ {0} = ensureInt16({0}); }}\n".format(js_arg)
         elif arg_type in ['Long', 'UnsignedLong']:
-          body += "  if (typeof arg%d == 'object') { arg%d = ensureInt32(arg%d); }\n" % (i, i, i)
+          body += "  if (typeof {0} == 'object') {{ {0} = ensureInt32({0}); }}\n".format(js_arg)
         elif arg_type == 'Float':
-          body += "  if (typeof arg%d == 'object') { arg%d = ensureFloat32(arg%d); }\n" % (i, i, i)
+          body += "  if (typeof {0} == 'object') {{ {0} = ensureFloat32({0}); }}\n".format(js_arg)
         elif arg_type == 'Double':
-          body += "  if (typeof arg%d == 'object') { arg%d = ensureFloat64(arg%d); }\n" % (i, i, i)
+          body += "  if (typeof {0} == 'object') {{ {0} = ensureFloat64({0}); }}\n".format(js_arg)
 
+  c_names = {}
   for i in range(min_args, max_args):
     c_names[i] = 'emscripten_bind_%s_%d' % (bindings_name, i)
-    body += '  if (arg%d === undefined) { %s%s(%s)%s%s }\n' % (i, call_prefix, '_' + c_names[i], ', '.join(pre_arg + args[:i]), call_postfix, '' if 'return ' in call_prefix else '; ' + (cache or ' ') + 'return')
+    body += '  if (%s === undefined) { %s%s(%s)%s%s }\n' % (args[i], call_prefix, '_' + c_names[i], ', '.join(pre_arg + args[:i]), call_postfix, '' if 'return ' in call_prefix else '; ' + (cache or ' ') + 'return')
   c_names[max_args] = 'emscripten_bind_%s_%d' % (bindings_name, max_args)
   body += '  %s%s(%s)%s;\n' % (call_prefix, '_' + c_names[max_args], ', '.join(pre_arg + args), call_postfix)
   if cache:
     body += '  ' + cache + '\n'
-  mid_js += [r'''function%s(%s) {
+  mid_js += [r'''%sfunction%s(%s) {
 %s
-};''' % ((' ' + func_name) if constructor else '', ', '.join(args), body[:-1])]
+};''' % (CONSTRUCTOR_CLOSURE_SUPPRESSIONS, (' ' + func_name) if constructor else '', ', '.join(args), body[:-1])]
 
   # C
 
-  for i in range(min_args, max_args+1):
+  for i in range(min_args, max_args + 1):
     raw = sigs.get(i)
-    if raw is None: continue
-    sig = map(full_typename, raw)
+    if raw is None:
+      continue
+    sig = list(map(full_typename, raw))
     if array_attribute:
-      sig = map(lambda x: x.replace('[]', ''), sig) # for arrays, ignore that this is an array - our get/set methods operate on the elements
+      sig = [x.replace('[]', '') for x in sig] # for arrays, ignore that this is an array - our get/set methods operate on the elements
 
-    c_arg_types = map(type_to_c, sig)
+    c_arg_types = list(map(type_to_c, sig))
 
-    normal_args = ', '.join(['%s arg%d' % (c_arg_types[j], j) for j in range(i)])
+    normal_args = ', '.join(['%s %s' % (c_arg_types[j], args[j]) for j in range(i)])
     if constructor:
       full_args = normal_args
     else:
       full_args = type_to_c(class_name, non_pointing=True) + '* self' + ('' if not normal_args else ', ' + normal_args)
-    call_args = ', '.join(['%sarg%d' % ('*' if raw[j].getExtendedAttribute('Ref') else '', j) for j in range(i)])
+    call_args = ', '.join(['%s%s' % ('*' if raw[j].getExtendedAttribute('Ref') else '', args[j]) for j in range(i)])
     if constructor:
       call = 'new ' + type_to_c(class_name, non_pointing=True)
       call += '(' + call_args + ')'
@@ -497,44 +557,46 @@ def render_function(class_name, func_name, sigs, return_type, non_pointer, copy,
       if class_name != func_scope:
         # this function comes from an ancestor class; for operators, we must cast it
         cast_self = 'dynamic_cast<' + type_to_c(func_scope) + '>(' + cast_self + ')'
+      maybe_deref = deref_if_nonpointer(raw[0])
       if '=' in operator:
-        call = '(*%s %s %sarg0)' % (cast_self, operator, '*' if sig[0] in interfaces else '')
+        call = '(*%s %s %s%s)' % (cast_self, operator, maybe_deref, args[0])
       elif operator == '[]':
-        call = '((*%s)[%sarg0])' % (cast_self, '*' if sig[0] in interfaces else '')
+        call = '((*%s)[%s%s])' % (cast_self, maybe_deref, args[0])
       else:
         raise Exception('unfamiliar operator ' + operator)
 
     pre = ''
 
-    basic_return = 'return ' if constructor or return_type is not 'Void' else ''
+    basic_return = 'return ' if constructor or return_type != 'Void' else ''
     return_prefix = basic_return
     return_postfix = ''
     if non_pointer:
-      return_prefix += '&';
+      return_prefix += '&'
     if copy:
       pre += '  static %s temp;\n' % type_to_c(return_type, non_pointing=True)
       return_prefix += '(temp = '
       return_postfix += ', &temp)'
 
     c_return_type = type_to_c(return_type)
+    maybe_const = 'const ' if const else ''
     mid_c += [r'''
 %s%s EMSCRIPTEN_KEEPALIVE %s(%s) {
 %s  %s%s%s;
 }
-''' % ('const ' if const else '', type_to_c(class_name) if constructor else c_return_type, c_names[i], full_args, pre, return_prefix, call, return_postfix)]
+''' % (maybe_const, type_to_c(class_name) if constructor else c_return_type, c_names[i], full_args, pre, return_prefix, call, return_postfix)]
 
     if not constructor:
       if i == max_args:
-        dec_args = ', '.join(map(lambda j: type_to_cdec(raw[j]) + ' arg' + str(j), range(i)))
-        js_call_args = ', '.join(['%sarg%d' % (('(int)' if sig[j] in interfaces else '') + ('&' if raw[j].getExtendedAttribute('Ref') or raw[j].getExtendedAttribute('Value') else ''), j) for j in range(i)])
+        dec_args = ', '.join([type_to_cdec(raw[j]) + ' ' + args[j] for j in range(i)])
+        js_call_args = ', '.join(['%s%s' % (('(int)' if sig[j] in interfaces else '') + take_addr_if_nonpointer(raw[j]), args[j]) for j in range(i)])
 
-        js_impl_methods += [r'''  %s %s(%s) {
+        js_impl_methods += [r'''  %s %s(%s) %s {
     %sEM_ASM_%s({
       var self = Module['getCache'](Module['%s'])[$0];
       if (!self.hasOwnProperty('%s')) throw 'a JSImplementation must implement all functions, you forgot %s::%s.';
       %sself['%s'](%s)%s;
     }, (int)this%s);
-  }''' % (c_return_type, func_name, dec_args,
+  }''' % (c_return_type, func_name, dec_args, maybe_const,
           basic_return, 'INT' if c_return_type not in C_FLOATS else 'DOUBLE',
           class_name,
           func_name, class_name, func_name,
@@ -545,9 +607,10 @@ def render_function(class_name, func_name, sigs, return_type, non_pointer, copy,
           (', ' if js_call_args else '') + js_call_args)]
 
 
-for name, interface in interfaces.iteritems():
+for name, interface in interfaces.items():
   js_impl = interface.getExtendedAttribute('JSImplementation')
-  if not js_impl: continue
+  if not js_impl:
+    continue
   implements[name] = [js_impl[0]]
 
 # Compute the height in the inheritance tree of each node. Note that the order of interation
@@ -559,7 +622,7 @@ for name, interface in interfaces.iteritems():
 # that invariant. Further, the height of a node never decreases. Therefore, when the loop
 # finishes, all ancestors of a given node should have a larger height number than that node.
 nodeHeight = {}
-for child, parent in implements.iteritems():
+for child, parent in implements.items():
   parent = parent[0]
   while parent:
     nodeHeight[parent] = max(nodeHeight.get(parent, 0), nodeHeight.get(child, 0) + 1)
@@ -570,8 +633,7 @@ for child, parent in implements.iteritems():
     else:
       parent = None
 
-names = interfaces.keys()
-names.sort(lambda x, y: nodeHeight.get(y, 0) - nodeHeight.get(x, 0))
+names = sorted(interfaces.keys(), key=lambda x: nodeHeight.get(x, 0), reverse=True)
 
 for name in names:
   interface = interfaces[name]
@@ -583,7 +645,8 @@ for name in names:
   js_impl_methods = []
 
   cons = interface.getExtendedAttribute('Constructor')
-  if type(cons) == list: raise Exception('do not use "Constructor", instead create methods with the name of the interface')
+  if type(cons) == list:
+    raise Exception('do not use "Constructor", instead create methods with the name of the interface')
 
   js_impl = interface.getExtendedAttribute('JSImplementation')
   if js_impl:
@@ -591,17 +654,14 @@ for name in names:
 
   # Methods
 
-  seen_constructor = False # ensure a constructor, even for abstract base classes
-  for m in interface.members:
-    if m.identifier.name == name:
-      seen_constructor = True
-      break
-  if not seen_constructor:
-    mid_js += ['function %s() { throw "cannot construct a %s, no constructor in IDL" }\n' % (name, name)]
-    emit_constructor(name)
+  # Ensure a constructor even if one is not specified.
+  if not any(m.identifier.name == name for m in interface.members):
+    mid_js += ['%sfunction %s() { throw "cannot construct a %s, no constructor in IDL" }\n' % (CONSTRUCTOR_CLOSURE_SUPPRESSIONS, name, name)]
+    mid_js += build_constructor(name)
 
   for m in interface.members:
-    if not m.isMethod(): continue
+    if not m.isMethod():
+      continue
     constructor = m.identifier.name == name
     if not constructor:
       parent_constructor = False
@@ -622,7 +682,7 @@ for name in names:
         return_type = ret.name
       else:
         assert return_type == ret.name, 'overloads must have the same return type'
-      for i in range(len(args)+1):
+      for i in range(len(args) + 1):
         if i == len(args) or args[i].optional:
           assert i not in sigs, 'overloading must differentiate by # of arguments (cannot have two signatures that differ by types but not by length)'
           sigs[i] = args[:i]
@@ -636,16 +696,17 @@ for name in names:
                     const=m.getExtendedAttribute('Const'))
     mid_js += [';\n']
     if constructor:
-      emit_constructor(name)
+      mid_js += build_constructor(name)
 
   for m in interface.members:
-    if not m.isAttr(): continue
+    if not m.isAttr():
+      continue
     attr = m.identifier.name
 
     if m.type.isArray():
-      get_sigs = { 1: [Dummy({ 'type': WebIDL.BuiltinTypes[WebIDL.IDLBuiltinType.Types.long] })] }
-      set_sigs = { 2: [Dummy({ 'type': WebIDL.BuiltinTypes[WebIDL.IDLBuiltinType.Types.long] }),
-                       Dummy({ 'type': m.type })] }
+      get_sigs = {1: [Dummy({'type': WebIDL.BuiltinTypes[WebIDL.IDLBuiltinType.Types.long]})]}
+      set_sigs = {2: [Dummy({'type': WebIDL.BuiltinTypes[WebIDL.IDLBuiltinType.Types.long]}),
+                      Dummy({'type': m.type})]}
       get_call_content = take_addr_if_nonpointer(m) + 'self->' + attr + '[arg0]'
       set_call_content = 'self->' + attr + '[arg0] = ' + deref_if_nonpointer(m) + 'arg1'
       if m.getExtendedAttribute('BoundsChecked'):
@@ -653,8 +714,8 @@ for name in names:
         get_call_content = "(%s, %s)" % (bounds_check, get_call_content)
         set_call_content = "(%s, %s)" % (bounds_check, set_call_content)
     else:
-      get_sigs = { 0: [] }
-      set_sigs = { 1: [Dummy({ 'type': m.type })] }
+      get_sigs = {0: []}
+      set_sigs = {1: [Dummy({'type': m.type})]}
       get_call_content = take_addr_if_nonpointer(m) + 'self->' + attr
       set_call_content = 'self->' + attr + ' = ' + deref_if_nonpointer(m) + 'arg0'
 
@@ -672,7 +733,10 @@ for name in names:
                     const=m.getExtendedAttribute('Const'),
                     array_attribute=m.type.isArray())
 
-    if not m.readonly:
+    if m.readonly:
+      mid_js += [r'''
+    Object.defineProperty(%s.prototype, '%s', { get: %s.prototype.%s });''' % (name, attr, name, get_name)]
+    else:
       set_name = 'set_' + attr
       mid_js += [r'''
     %s.prototype['%s'] = %s.prototype.%s = ''' % (name, set_name, name, set_name)]
@@ -686,12 +750,14 @@ for name in names:
                       call_content=set_call_content,
                       const=m.getExtendedAttribute('Const'),
                       array_attribute=m.type.isArray())
+      mid_js += [r'''
+    Object.defineProperty(%s.prototype, '%s', { get: %s.prototype.%s, set: %s.prototype.%s });''' % (name, attr, name, get_name, name, set_name)]
 
   if not interface.getExtendedAttribute('NoDelete'):
     mid_js += [r'''
   %s.prototype['__destroy__'] = %s.prototype.__destroy__ = ''' % (name, name)]
     render_function(name,
-                    '__destroy__', { 0: [] }, 'Void',
+                    '__destroy__', {0: []}, 'Void',
                     None,
                     None,
                     None,
@@ -711,50 +777,51 @@ public:
 
 deferred_js = []
 
-for name, enum in enums.iteritems():
+for name, enum in enums.items():
   mid_c += ['\n// ' + name + '\n']
   deferred_js += ['\n', '// ' + name + '\n']
   for value in enum.values():
     function_id = "%s_%s" % (name, value.split('::')[-1])
-    mid_c += [r'''%s EMSCRIPTEN_KEEPALIVE emscripten_enum_%s() {
+    function_id = 'emscripten_enum_%s' % function_id
+    mid_c += [r'''%s EMSCRIPTEN_KEEPALIVE %s() {
   return %s;
 }
 ''' % (name, function_id, value)]
     symbols = value.split('::')
     if len(symbols) == 1:
       identifier = symbols[0]
-      deferred_js += ["Module['%s'] = _emscripten_enum_%s();\n" % (identifier, function_id)]
+      deferred_js += ["Module['%s'] = _%s();\n" % (identifier, function_id)]
     elif len(symbols) == 2:
       [namespace, identifier] = symbols
       if namespace in interfaces:
         # namespace is a class
-        deferred_js += ["Module['%s']['%s'] = _emscripten_enum_%s();\n" % \
-                  (namespace, identifier, function_id)]
+        deferred_js += ["Module['%s']['%s'] = _%s();\n" % (namespace, identifier, function_id)]
       else:
         # namespace is a namespace, so the enums get collapsed into the top level namespace.
-        deferred_js += ["Module['%s'] = _emscripten_enum_%s();\n" % (identifier, function_id)]
+        deferred_js += ["Module['%s'] = _%s();\n" % (identifier, function_id)]
     else:
       raise Exception("Illegal enum value %s" % value)
 
 mid_c += ['\n}\n\n']
-mid_js += ['''
+if len(deferred_js):
+  mid_js += ['''
 (function() {
   function setupEnums() {
     %s
   }
-  if (Module['calledRun']) setupEnums();
+  if (runtimeInitialized) setupEnums();
   else addOnPreMain(setupEnums);
 })();
 ''' % '\n    '.join(deferred_js)]
 
 # Write
 
-c = open(output_base + '.cpp', 'w')
-for x in pre_c: c.write(x)
-for x in mid_c: c.write(x)
-c.close()
-
-js = open(output_base + '.js', 'w')
-for x in mid_js: js.write(x)
-js.close()
+with open(output_base + '.cpp', 'w') as c:
+  for x in pre_c:
+    c.write(x)
+  for x in mid_c:
+    c.write(x)
 
+with open(output_base + '.js', 'w') as js:
+  for x in mid_js:
+    js.write(x)
